package com.mayank.training.jbpm.services;

import java.util.List;
import java.util.Map;
import java.util.Properties;

import javax.persistence.EntityManagerFactory;
import javax.persistence.Persistence;

import org.apache.log4j.Logger;
import org.drools.KnowledgeBase;
import org.drools.KnowledgeBaseFactory;
import org.drools.builder.KnowledgeBuilder;
import org.drools.builder.KnowledgeBuilderError;
import org.drools.builder.KnowledgeBuilderFactory;
import org.drools.builder.ResourceType;
import org.drools.io.ResourceFactory;
import org.drools.persistence.jpa.JPAKnowledgeService;
import org.drools.runtime.Environment;
import org.drools.runtime.EnvironmentName;
import org.drools.runtime.KnowledgeSessionConfiguration;
import org.drools.runtime.StatefulKnowledgeSession;
import org.jbpm.process.audit.JPAProcessInstanceDbLog;
import org.jbpm.process.audit.JPAWorkingMemoryDbLogger;
import org.jbpm.process.workitem.wsht.LocalHTWorkItemHandler;
import org.jbpm.task.identity.DefaultUserGroupCallbackImpl;
import org.jbpm.task.identity.UserGroupCallbackManager;
import org.jbpm.task.service.TaskService;
import org.jbpm.task.service.local.LocalTaskService;
import org.jbpm.task.utils.OnErrorAction;
import org.springframework.transaction.annotation.Transactional;

import bitronix.tm.TransactionManagerServices;

@Transactional
public abstract class JbpmService {
	
	private static final Logger logger_ = Logger.getLogger(JbpmService.class);
	private final Environment env;
	private final EntityManagerFactory emf;
	private final KnowledgeSessionConfiguration conf;
	protected final KnowledgeBase kbase;
	protected LocalTaskService localTaskService;
	private StatefulKnowledgeSession ksession;

	
	public JbpmService(List<String> resources, LocalTaskService taskService, EntityManagerFactory emf) throws Exception{
		 this.emf = emf;
	     this.env = KnowledgeBaseFactory.newEnvironment();
	     this.env.set(EnvironmentName.ENTITY_MANAGER_FACTORY, emf);
	     logger_.info("Will use:"+ TransactionManagerServices.getTransactionManager().toString() + ":for transactions");
	     this.env.set(EnvironmentName.TRANSACTION_MANAGER, TransactionManagerServices.getTransactionManager());
	     //this.env.set(EnvironmentName.TRANSACTION, TransactionManagerServices.getTransactionManager());
	     this.conf = KnowledgeBaseFactory.newKnowledgeSessionConfiguration();
	     this.kbase = createKnowledgeBase(resources);
	     this.localTaskService = taskService;
	}
	
	
	private void registerProcessHistroyLogger(StatefulKnowledgeSession session){
		new JPAWorkingMemoryDbLogger(session);
    	JPAProcessInstanceDbLog log = new JPAProcessInstanceDbLog(session.getEnvironment());
	}
	
	private void registerWorkItemHandlers(StatefulKnowledgeSession ksession){
		if(null != ksession){
			UserGroupCallbackManager.getInstance().setCallback(new DefaultUserGroupCallbackImpl("classpath:/usergroups.properties"));
			LocalHTWorkItemHandler humanTaskHandler = new LocalHTWorkItemHandler(localTaskService,ksession,OnErrorAction.RETHROW);
			humanTaskHandler.setLocal(true);
			humanTaskHandler.connect();
			ksession.getWorkItemManager().registerWorkItemHandler("Human Task", humanTaskHandler);
		}
	}
	
    private KnowledgeBase createKnowledgeBase(String... process) {
        KnowledgeBuilder kbuilder = KnowledgeBuilderFactory.newKnowledgeBuilder();
        for (String p: process) {
            kbuilder.add(ResourceFactory.newClassPathResource(p), ResourceType.BPMN2);
        }
        if (kbuilder.hasErrors()) {
            if (kbuilder.getErrors().size() > 0) {
                boolean errors = false;
                for (KnowledgeBuilderError error : kbuilder.getErrors()) {
                    logger_.error(error.toString());
                    errors = true;
                }
                if(errors)
                	throw new RuntimeException("Could not build KnwoeldgeBase...");
            }
        }
        return kbuilder.newKnowledgeBase();
    }
	
	private KnowledgeBase createKnowledgeBase(List<String> resources) throws Exception {
		KnowledgeBuilder kbuilder = KnowledgeBuilderFactory.newKnowledgeBuilder();
		for (String resource: resources) {
			if(resource.toLowerCase().contains("drl"))
				kbuilder.add(ResourceFactory.newClassPathResource(resource), ResourceType.DRL);
			if(resource.toLowerCase().contains("bpmn"))
				kbuilder.add(ResourceFactory.newClassPathResource(resource), ResourceType.BPMN2);
		}
		return kbuilder.newKnowledgeBase();
	}
	
	protected StatefulKnowledgeSession createKnowledgeSession() {
        ksession = JPAKnowledgeService.newStatefulKnowledgeSession(kbase, conf, env);
        registerWorkItemHandlers(ksession);
        registerProcessHistroyLogger(ksession);
		return ksession;
	}
	
	protected StatefulKnowledgeSession loadKnowledgeSession(Long sessionId){
		StatefulKnowledgeSession ksession = JPAKnowledgeService.loadStatefulKnowledgeSession(sessionId.intValue(), kbase, conf, env);
	    registerWorkItemHandlers(ksession);
	    registerProcessHistroyLogger(ksession);
		return ksession;
	}
	
	/**
	 * when loading a persisted session/processId verify jbpm db and training db has
	 * same thing in place.
	 * @param sessionId
	 * @param processId
	 * @return
	 */
	protected boolean isSessionProcessInstanceMappingCorrect(Long sessionId, Long processId){
		StatefulKnowledgeSession session = loadKnowledgeSession(sessionId);
		if (null != session.getProcessInstance(processId))
			return true;
		return false;
	}
	
	public StatefulKnowledgeSession getSession(){
		return this.ksession;
	}
	
    private StatefulKnowledgeSession createKnowledgeSession(String... process) {
			KnowledgeBase kbase = createKnowledgeBase(process);
			return createKnowledgeSession();
	}
	
}
